/* Copyright (C) 1989, 2000 Aladdin Enterprises.  All rights reserved.
  
  This file is part of AFPL Ghostscript.
  
  AFPL Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author or
  distributor accepts any responsibility for the consequences of using it, or
  for whether it serves any particular purpose or works at all, unless he or
  she says so in writing.  Refer to the Aladdin Free Public License (the
  "License") for full details.
  
  Every copy of AFPL Ghostscript must include a copy of the License, normally
  in a plain ASCII text file named PUBLIC.  The License grants you the right
  to copy, modify and redistribute AFPL Ghostscript, but only under certain
  conditions described in the License.  Among other things, the License
  requires that the copyright notice and this notice be preserved on all
  copies.
*/

/*$Id: gsplugin.c,v 1.0 2001/04/06 22:23:44 raph Exp $ */
/* Plugin manager. */
#include "gx.h"
#include "gp.h"
#include "gsplugin.h"
#include "gzstate.h"

/*  Since on some platforms DLLs must not load twice,
    we need to maintain instance-independent (i.e. static)
    list of loaded DLLs. Therefore plugin manager's data
    are split into static data and instance data.
    In multithreading environment the maintanence of
    static data requires mutex - to be fixed later
    (see comments in code).

    In this revision we allocate fixed size arrays
    for lists. To be fixed later with linked lists and
    dymanical memory allocation for them.
    Need to decide, which memory pool is
    appropriate for this, as well as resolve
    the order of initialization of memory pools.

    Rather clients may load plugin DLLs independently on this manager,
    plugin directories used by clients and by the manager 
    must not intersect. Otherwice a DLL may be load twice,
    and we can't guarantee safe unloading it.

    Now plugin manager is a part of graphics library,
    because we don't have a separate product for it.

    This module mixes code of general-purpose plugin manager, and
    specific code for graphics library, which is a client of
    the plugin manager. Some day we should split this module into
    into 2 parts.
*/

private int version_major_static = 1;

typedef struct gs_plugin_registration_request_s gs_plugin_registration_request;
typedef struct gs_plugin_manager_statics_s gs_plugin_manager_statics;
typedef struct gs_plugin_manager_instance_s gs_plugin_manager_instance;
#ifndef gs_instance_DEFINED
#  define gs_instance_DEFINED
typedef struct gs_instance_s gs_instance;
#endif

struct gs_plugin_registration_request_s {
    int version_major;
    char *interface_type;
    void *client_id;
    void *client_data;
    gs_plugin_registration_callback do_register;
    gs_plugin_registration_callback unregister;
};

struct gs_plugin_holder_s {
    gs_plugin_interface *I;
    void *client_id;
    void *hmodule;
};

struct gs_plugin_manager_statics_s {
    int holder_count;           /* # of plugin holders */
    struct gs_plugin_holder_s holder[20]; /* Holders of loaded plugins. */
};

struct gs_plugin_manager_instance_s { /* instance-dependent data */
    int request_count;          /* # of requested interface types */
    gs_plugin_registration_request request[5];
};


struct gs_instance_s { /* Graphics library instance - reserved for further improvements about reentrancy. */
    gs_plugin_manager_instance plmi;
    int interface_count;        /* # of interfaces installed to graphics library */
    gs_plugin_interface *I[11]; /* Interfaces installed to graphics library, zero-terminated array */
};

/* Specific functions for graphics library : */

private gs_plugin_retcode register_plugin_for_gxlib(void *host, gs_plugin_interface *I)
{   gs_instance *gsi = (gs_instance *)host;
    if (gsi->interface_count >= count_of(gsi->I) - 1)
        return gs_plugin__rc_excess;
    gsi->I[gsi->interface_count] = I;
    gsi->interface_count ++;
    gsi->I[gsi->interface_count] = 0;
    return gs_plugin__rc_success;
}

private gs_plugin_retcode unregister_plugin_for_gxlib(void *host, gs_plugin_interface *I)
{   gs_instance *gsi = (gs_instance *)host;
    int k , j = 0;
    for (k = 0; k < gsi->interface_count; k++)
        if(gsi->I[k] != I)
           gsi->I[j++] = gsi->I[k];
    gsi->interface_count = j;
    return gs_plugin__rc_success;
}

private gs_plugin_manager_statics plms = {
    0 /* holder_count */
};

#define GX_INTERFACE_TYPE "GX"

private const char *gx_interface_type = GX_INTERFACE_TYPE;


private gs_instance gli0 = { /* Define this statically until fixing gslib reentrancy. */
    {   /* plmi */
        1,  /* request_count */
        {   /* plugin_request array */
            {   /* plugin_request[0] - own needs of Graphics Library */
                1,     /* version_major */
                GX_INTERFACE_TYPE,  /* interface_type. */
                &gli0, /* client_id */
                &gli0, /* client_data */
                register_plugin_for_gxlib, /* do_register */
                unregister_plugin_for_gxlib /* unregister */
            }
        }
    },
    0, /* interface_count */
    {0} /* I */
};

gs_plugin_retcode gs_unregister_gxlib_as_plugin_client(gs_instance *gli)
{   return gs_unregister_plugin_client(gli, &gli0);
}

gs_plugin_interface ** gs_get_installed_plugins(gs_instance *gli)
{   gli = &gli0; /* Temporary solution befor fixing the reentrancy of graphics library. */
    return gli->I;
}



/* Function members of the plugin manager, being independent on graphics library : 
   (fixme : splitting the module, replace gs_instance argument with gs_plugin_manager_instance argument)
*/

gs_plugin_retcode gs_add_plugin_registration_request(gs_instance *gli, int version_major, char *interface_type, void *client_id, void *client_data, gs_plugin_registration_callback do_register, gs_plugin_registration_callback unregister)
{   gs_plugin_manager_instance *plmi = &gli0.plmi;
    if (plmi->request_count >= count_of(plmi->request))
        return gs_plugin__rc_excess;
    plmi->request[plmi->request_count].version_major = version_major;
    plmi->request[plmi->request_count].interface_type = interface_type;
    plmi->request[plmi->request_count].client_id = client_id;
    plmi->request[plmi->request_count].client_data = client_data;
    plmi->request[plmi->request_count].do_register = do_register;
    plmi->request[plmi->request_count].unregister = unregister;
    plmi->request_count ++;
    return gs_plugin__rc_success;
    /*  fixme : if request comes after some plugins already held,
        need to obtain more interfaces from them, and register them with
        the request. This may be done with adding "plugin instance id" to 
        gs_plugin_holder, which is the initial interface for a plugin instance.
        Currently not supported, because for involving this facility 
        GS DLL API needs to be fixed about the GS initialization procedure.
    */
}

private gs_plugin_retcode hold_interface(gs_instance *gli, gs_plugin_interface *I, void *client_id, void *hmodule)
{   /* fixme: lock mutex here. */
    if (plms.holder_count >= count_of(plms.holder))
        return gs_plugin__rc_excess;
    plms.holder[plms.holder_count].I = I;
    plms.holder[plms.holder_count].client_id = client_id;
    plms.holder[plms.holder_count].hmodule = hmodule;
    plms.holder_count ++;
    /* fixme: unlock mutex here. */
    return gs_plugin__rc_success;
}

private gs_plugin_retcode try_unload_module(gs_instance *gli, PFN_get_factories gf, void *hmodule)
{   gs_plugin_factory **factory_list = gf(version_major_static);
    gs_plugin_retcode code = gs_plugin__rc_success, code1;
    int k;
    for (k = 0; factory_list[k]; k++) {
        gs_plugin_factory *f = factory_list[k];
        code1 = f->can_close(f);
        if (code1 == gs_plugin__rc_busy)
            return code;
        if (code1 != gs_plugin__rc_success) {
            /* We better keep tha module than unload it with possible further crash :*/
            if (code != gs_plugin__rc_success)
                return code;
            return code1;
        }
    }
    for (k = 0; factory_list[k]; k++) {
        gs_plugin_factory *f = factory_list[k];
        code1 = f->close(f);
        if (code == gs_plugin__rc_success)
            code = code1;
    }
    gp_dlclose(hmodule);
    return code;
}

private gs_plugin_retcode unhold_interfaces_(gs_instance *gli, gs_plugin_interface *I, void *client_id, bool unload_even_on_error)
{   gs_plugin_manager_instance *plmi = &gli0.plmi;
    int k, j = 0, i;
    gs_plugin_retcode code = gs_plugin__rc_success, code1;
    gs_plugin_interface *I1;
    for (k = 0; k < plms.holder_count; k++)
        if ((I == 0 || plms.holder[k].I == I) && 
                (client_id == 0 || plms.holder[k].client_id == client_id)) {
            void *hmodule;
            PFN_get_factories gf;
            gs_plugin_factory *f;
            I1 = plms.holder[k].I;
            for (i = 0; i<plmi->request_count; i++)
                if (plmi->request[i].client_id == client_id) {
                    code1 = plmi->request[i].unregister(plmi->request[i].client_data, I1);
                    if (code == gs_plugin__rc_success)
                        code = code1;
                }
            f = I1->factory;
            hmodule = plms.holder[k].hmodule;
            *(void **)&gf = gp_dlsym(hmodule, "get_factories");
            code1 = I1->release_interface(I1);
            if (code == gs_plugin__rc_success)
                code = code1;
            code1 = f->can_close(f);
            if (code1 == gs_plugin__rc_busy)
                continue;
            if (code == gs_plugin__rc_success)
                code = code1;
            code1 = try_unload_module(gli, gf, hmodule);
            if (code == gs_plugin__rc_success)
                code = code1;
        } else
           plms.holder[j++] = plms.holder[k];
    plms.holder_count = j;
    return code;
}

private gs_plugin_retcode unhold_interfaces(gs_instance *gli, gs_plugin_interface *I, void *client_id, bool unload_even_on_error)
{   gs_plugin_retcode code;
    /* fixme: lock mutex here. */
    code = unhold_interfaces_(gli, I, client_id, unload_even_on_error);
    /* fixme: unlock mutex here. */
    return code;
}

private gs_plugin_retcode gs_register_plugin_instance(gs_instance *gli, gs_plugin_interface *I0, const char *interface_type, void *hmodule, void *client_id)
{   /* I0 to be held or released. */
    gs_plugin_manager_instance *plmi = &gli0.plmi;
    gs_plugin_interface *I;
    int k, hold_count = 0;
    gs_plugin_retcode code = gs_plugin__rc_success, code1;
    for (k = 0; k < plmi->request_count; k++) {
        if (!strcmp(interface_type, plmi->request[k].interface_type)) {
            I = I0;
        } else
            I = I0->get_interface(I0, plmi->request[k].interface_type, plmi->request[k].version_major);
        if (I != 0) {
            code1 = hold_interface(gli, I, plmi->request[k].client_id, hmodule);
            if (code == gs_plugin__rc_success)
                code = code1;
            if (I == I0)
                hold_count++;
            code1 = plmi->request[k].do_register(plmi->request[k].client_data, I);
            if (code == gs_plugin__rc_success)
                code = code1;
        }
    }
    if (hold_count == 0) {
        code1 = I0->release_interface(I0);
        if (code == gs_plugin__rc_success)
            code = code1;
    }
    return code;
}

gs_plugin_retcode gs_register_plugin(gs_instance *gli, gs_plugin_interface *I0, const char *interface_type, void *client_id)
{   /* Exposed for gxlib clients. */
    return gs_register_plugin_instance(gli, I0, interface_type, 0, client_id);
}

gs_plugin_retcode gs_unregister_plugin_client(gs_instance *gli, void *client_id)
{   /* Exposed for gxlib clients. */
    return unhold_interfaces(gli, 0, client_id, true);
}

private gs_plugin_retcode gs_register_plugin_(gs_instance *gli, PFN_get_factories gf, void *hmodule, void *client_id)
{   /* May be called from client with hmodule == 0. */
    /* Must either hold an interface or release the module (if hmodule != 0). */
    int k;
    gs_plugin_retcode code = gs_plugin__rc_success, code1;
    gs_plugin_factory **factory_list = gf(version_major_static);
    gs_plugin_interface *I;
    if (!factory_list[0]) {
        if (hmodule != 0)
            gp_dlclose(hmodule);
        return gs_plugin__rc_success;
    }
    for (k = 0; factory_list[k] && code == gs_plugin__rc_success; k++) {
        gs_plugin_factory *factory = factory_list[k];
        char *s0 = "Failed installing plugin : ", *s1 = ", version ";
        gs_plugin_factory *f = factory;
        code = factory->make_instance(factory, gx_interface_type, version_major_static, &I);
        if (code == gs_plugin__rc_success) {
            f = I->factory;
            code1 = gs_register_plugin_instance(gli, I, gx_interface_type, hmodule, client_id);
            if (code1 != gs_plugin__rc_success)
                code = code1;
            else
                s0 = "Installed plugin : ";
        }
        dlprintf(s0);
        dlprintf(f->description);
        dlprintf(s1);
        dlprintf2("%d.%d\n", f->version_major, f->version_minor);
    }
    if (hmodule == 0)
        return code;
    /* We cannot know whether something was held, so check all factories for closing : */
    return try_unload_module(gli, gf, hmodule);
}

gs_plugin_retcode gs_register_plugin_set(gs_instance *gli, PFN_get_factories gf, void *client_id)
{   /* Exposed for gxlib clients. */
    return gs_register_plugin_(gli, gf, 0, client_id);
}

gs_plugin_retcode gs_load_plugins(gs_instance *gli, gs_memory_t *mem, const char *dir_path, void *client_id)
{   PFN_get_factories pfn_get_factories;
    file_enum * pfen;
    gs_plugin_retcode code = gs_plugin__rc_success, code1 = code, code2;
    int L = strlen(dir_path) + gp_file_name_sizeof;
    char *buf;
    if (dir_path == 0 || dir_path[0] == 0)
        return gs_plugin__rc_success;
    buf = (char *)gs_alloc_string(mem, L, "plugin path");
    strcpy(buf, dir_path);
    strcat(buf, gp_file_name_concat_string(buf, strlen(buf), 
           gp_file_name_generic_wildcard(), strlen(gp_file_name_generic_wildcard())));
    strcat(buf, gp_file_name_generic_wildcard()); 
    pfen = gp_enumerate_files_init(buf, strlen(buf), mem);
    if (pfen == 0)
        code = gs_plugin__rc_memory;
    else
        while (code2 = gp_enumerate_files_next(pfen, buf, L - 1),
               code2 != 0 && code2 != ~ (uint)0) {
            void *hmodule;
	    buf[code2] = 0;
            if (!gp_is_dll(buf))
                continue;
            hmodule = gp_dlopen(buf);
            pfn_get_factories = 0;
            if (hmodule)
                *(void **)&pfn_get_factories = gp_dlsym(hmodule, "get_factories");
            if (pfn_get_factories == 0) {
                code = gs_plugin__rc_export;
                dprintf1("Can't register %s as plugin: not a DLL or no 'get_factories'", buf);
                gp_dlclose(hmodule);
                continue;
            }
            code1 = gs_register_plugin_(gli, pfn_get_factories, hmodule, client_id);
            if (code != gs_plugin__rc_success) {
                dprintf2("DLL %s can't register as plugin, return code = %d.\n", buf, code1);
                if (code != gs_plugin__rc_success)
                    code = code1;
            }
        }
    gs_free_string(mem, (byte *)buf, L, "plugin path");
    return code;
}





